using System;
using System.Data;
using System.Configuration;
using System.Web;
using System.Web.Security;
using System.Web.UI;
using System.Web.UI.WebControls;
using System.Web.UI.WebControls.WebParts;
using System.Web.UI.HtmlControls;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using Microsoft.Win32;
using System.Globalization;

namespace TimeZoneLibrary
{
    [StructLayout(LayoutKind.Sequential)]
    struct SYSTEMTIME
    {
        public ushort wYear;
        public ushort wMonth;
        public ushort wDayOfWeek;
        public ushort wDay;
        public ushort wHour;
        public ushort wMinute;
        public ushort wSecond;
        public ushort wMilliseconds;


        public void SetInfo(byte[] info)
        {
            if (info.Length != Marshal.SizeOf(this))
            {
                throw new ArgumentException("Information size is incorrect", "info");
            }
            else
            {
                this.wYear = BitConverter.ToUInt16(info, 0);
                this.wMonth = BitConverter.ToUInt16(info, 2);
                this.wDayOfWeek = BitConverter.ToUInt16(info, 4);
                this.wDay = BitConverter.ToUInt16(info, 6);
                this.wHour = BitConverter.ToUInt16(info, 8);
                this.wMinute = BitConverter.ToUInt16(info, 10);
                this.wSecond = BitConverter.ToUInt16(info, 12);
                this.wMilliseconds = BitConverter.ToUInt16(info, 14);
            }
        }


        public override bool Equals(object obj)
        {
            if (this.GetType() == obj.GetType())
            {
                try
                {
                    SYSTEMTIME objSt = (SYSTEMTIME)obj;
                    if (this.wYear != objSt.wYear
                        || this.wMonth != objSt.wMonth
                        || this.wDayOfWeek != objSt.wDayOfWeek
                        || this.wDay != objSt.wDay
                        || this.wHour != objSt.wHour
                        || this.wMinute != objSt.wMinute
                        || this.wSecond != objSt.wSecond
                        || this.wMilliseconds != objSt.wMilliseconds)
                        return false;
                    else
                        return true;
                }
                catch
                {
                    return false;
                }
            }
            else
                return false;
        }


        public override int GetHashCode()
        {
            return base.GetHashCode();
        }
    }

    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    struct TimeZoneInformation
    {
        public int bias;
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 32)]
        public string standardName;
        public SYSTEMTIME standardDate;
        public int standardBias;
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 32)]
        public string daylightName;
        public SYSTEMTIME daylightDate;
        public int daylightBias;


        public void SetBytes(byte[] info)
        {
            if (info.Length != 44)
                throw new ArgumentException("Information size is incorrect", "info");
            else
            {
                this.bias = BitConverter.ToInt32(info, 0);
                this.standardBias = BitConverter.ToInt32(info, 4);
                this.daylightBias = BitConverter.ToInt32(info, 8);
                byte[] helper = new byte[16];
                Array.Copy(info, 12, helper, 0, 16);
                this.standardDate.SetInfo(helper);
                Array.Copy(info, 28, helper, 0, 16);
                this.daylightDate.SetInfo(helper);
            }
        }

        public override bool Equals(object obj)
        {
            if (this.GetType() != obj.GetType())
            {
                try
                {
                    TimeZoneInformation objTzi = (TimeZoneInformation)obj;
                    if (this.bias != objTzi.bias
                        || this.daylightBias != objTzi.daylightBias
                        || this.daylightName != objTzi.daylightName
                        || this.standardBias != objTzi.standardBias
                        || this.standardName != objTzi.standardName
                        || !this.daylightDate.Equals(objTzi.daylightDate)
                        || !this.standardDate.Equals(objTzi.standardDate)
                        )
                        return false;
                    else
                        return true;
                }
                catch
                {
                    return false;
                }
            }
            else
                return false;
        }

        public override int GetHashCode()
        {
            return base.GetHashCode();
        }
    }

    public class TimeZoneInfo : IComparer<TimeZoneInfo>
    {
        private string _id;
        private TimeZoneInformation _tzi;
        private string _displayName;

        public string Id
        {
            get
            {
                Refresh();
                return this._id;
            }
        }

        public string DisplayName
        {

            get
            {
                Refresh();
                return _displayName;
            }
        }

        public string StandardName
        {
            get
            {
                Refresh();
                return this._tzi.standardName;
            }
        }

        public TimeSpan StandardUtcOffset
        {
            get
            {
                Refresh();
                return new TimeSpan(0, -this._tzi.bias, 0);
            }
        }

        public string DaylightName(DateTime dtDate)
        {
            Refresh();
            if (this.GetDaylightChanges(dtDate.Year).Delta == TimeSpan.Zero)
                return this._tzi.standardName;
            else
                return this._tzi.daylightName;
        }

        public TimeSpan GetCurrentUtcOffset(DateTime dtDate)
        {
            if (this.IsDaylightSavingTime(dtDate))
                return new TimeSpan(0, -(this._tzi.bias + this._tzi.daylightBias), 0);
            else
                return new TimeSpan(0, -this._tzi.bias, 0);
        }

        public static TimeZoneInfo[] GetTimeZones()
        {
            List<TimeZoneInfo> tzInfos = new List<TimeZoneInfo>();
            RegistryKey key = GetTimeZoneRegistryKey();
            foreach (string zoneName in key.GetSubKeyNames())
            {
                TimeZoneInfo tzi = new TimeZoneInfo();
                tzi._id = zoneName;
                tzi.SetValues();
                tzInfos.Add(tzi);
            }
            TimeZoneInfo.Sort(tzInfos);
            return tzInfos.ToArray();
        }

        public DaylightTime GetDaylightChanges(int year)
        {
            TimeZoneInformation tzi = new TimeZoneInformation();
            RegistryKey subKey1 = GetTimeZoneRegistryDynamicSubKey(this);
            if (subKey1 != null)
            {
                if (Array.IndexOf(subKey1.GetValueNames(), year.ToString()) != -1)
                    tzi.SetBytes(subKey1.GetValue(year.ToString()) as byte[]);
                else
                {
                    Refresh();
                    tzi = this._tzi;
                }
            }
            else
            {
                Refresh();
                tzi = this._tzi;
            }

            DateTime dStart;
            DateTime dEnd;
            dStart = this.GetStartDate(tzi, year);
            dEnd = this.GetEndDate(tzi, year);
            if (dStart != DateTime.MinValue && dEnd != DateTime.MinValue)
                return new DaylightTime(dStart, dEnd, new TimeSpan(0, -this._tzi.daylightBias, 0));
            else
                return new DaylightTime(dStart, dEnd, new TimeSpan(0, 0, 0));
        }

        public bool IsDaylightSavingTime(DateTime dtDate)
        {
            DateTime dUtcTime = dtDate; //dtDate.AddMinutes(-this._tzi.bias);
            DateTime sUtcTime = dtDate.AddMinutes(-this._tzi.daylightBias); //dtDate.AddMinutes(-(this._tzi.bias + this._tzi.daylightBias));
            DaylightTime dt;

            if (this._tzi.daylightDate.wMonth <= this._tzi.standardDate.wMonth)
            {
                //Daylight saving time starts and ends in the same year
                dt = this.GetDaylightChanges(dUtcTime.Year);
                if (dt.Delta != TimeSpan.Zero)
                {
                    if (dUtcTime >= dt.Start && sUtcTime < dt.End)
                        return true;
                    else
                        return false;
                }
            }
            else
            {
                //Daylight saving time starts and ends in diferent years
                dt = this.GetDaylightChanges(sUtcTime.Year);
                if (dt.Delta != TimeSpan.Zero)
                {
                    if (dUtcTime < dt.Start && sUtcTime >= dt.End)
                        return false;
                    else
                        return true;
                }
            }
            return false;
        }

        private DateTime GetStartDate(TimeZoneInformation tzi, int year)
        {
            if (tzi.daylightDate.wMonth != 0)
            {
                if (tzi.daylightDate.wYear == 0)
                    return this.CreateDate(year, tzi.daylightDate.wMonth, tzi.daylightDate.wDay, tzi.daylightDate.wDayOfWeek, tzi.daylightDate.wHour, tzi.daylightDate.wMinute, tzi.daylightDate.wSecond, tzi.daylightDate.wMilliseconds);
                else
                    return new DateTime(tzi.daylightDate.wYear, tzi.daylightDate.wMonth, tzi.daylightDate.wDay, tzi.daylightDate.wHour, tzi.daylightDate.wMinute, tzi.daylightDate.wSecond, tzi.daylightDate.wMilliseconds, DateTimeKind.Local);
            }
            else
            {
                return DateTime.MinValue;
            }
        }

        private DateTime CreateDate(int wYear, int wMonth, int wDay, int wDayOfWeek, int wHour, int wMinute, int wSecond, int wMilliseconds)
        {
            if (wDay < 1 || wDay > 5)
                throw new ArgumentOutOfRangeException(@"wDat", wDay, @"The value is out of acceptable range (1 to 5).");
            if (wDayOfWeek < 0 || wDayOfWeek > 6)
                throw new ArgumentOutOfRangeException(@"wDayOfWeek", wDayOfWeek, @"The value is out of acceptable range (0 to 6).");
            int daysInMonth = DateTime.DaysInMonth(wYear, wMonth);
            int fDayOfWeek = (int)new DateTime(wYear, wMonth, 1).DayOfWeek;
            int occurre = 1;
            int day = 1;
            if (fDayOfWeek != wDayOfWeek)
            {
                if (wDayOfWeek == 0)
                    day += 7 - fDayOfWeek;
                else
                {
                    if (wDayOfWeek > fDayOfWeek)
                    {
                        day += wDayOfWeek - fDayOfWeek;
                    }
                    else if (wDayOfWeek < fDayOfWeek)
                    {
                        day = wDayOfWeek + fDayOfWeek;
                    }
                }
            }
            while (occurre < wDayOfWeek && day <= daysInMonth - 7)
            {
                day += 7;
                occurre += 1;
            }
            return new DateTime(wYear, wMonth, day, wHour, wMinute, wSecond, wMilliseconds, DateTimeKind.Local);
        }

        private DateTime GetEndDate(TimeZoneInformation tzi, int year)
        {
            if (tzi.standardDate.wMonth != 0)
            {
                if (tzi.standardDate.wYear == 0)
                    return this.CreateDate(year, tzi.standardDate.wMonth, tzi.standardDate.wDay, tzi.standardDate.wDayOfWeek, tzi.standardDate.wHour, tzi.standardDate.wMinute, tzi.standardDate.wSecond, tzi.standardDate.wMilliseconds);
                else
                    return new DateTime(tzi.standardDate.wYear, tzi.standardDate.wMonth, tzi.standardDate.wDay, tzi.standardDate.wHour, tzi.standardDate.wMinute, tzi.standardDate.wSecond, tzi.standardDate.wMilliseconds, DateTimeKind.Local);
            }
            else
            {
                return DateTime.MinValue;
            }
        }

        public void Refresh()
        {
            this.SetValues();
        }

        private void SetValues()
        {
            RegistryKey subKey = GetTimeZoneRegistrySubKey(this._id);
            this._displayName = subKey.GetValue("Display").ToString();
            this._tzi.daylightName = subKey.GetValue("Dlt").ToString();
            this._tzi.standardName = subKey.GetValue("Std").ToString();
            this._tzi.SetBytes(subKey.GetValue("Tzi") as byte[]);
        }

        private static RegistryKey GetTimeZoneRegistryKey()
        {
            RegistryKey key = Registry.LocalMachine.OpenSubKey(@"SOFTWARE\Microsoft\Windows NT\CurrentVersion\Time Zones", false);
            if (key == null)
                throw new KeyNotFoundException(@"Cannot find the windows registry key (Time Zone).");
            else
                return key;
        }

        private static RegistryKey GetTimeZoneRegistrySubKey(string zoneId)
        {
            RegistryKey key = GetTimeZoneRegistryKey();
            RegistryKey subKey = key.OpenSubKey(zoneId, false);
            if (subKey == null)
                throw new Exception("Unknown time zone.");
            else
                return subKey;
        }

        private static RegistryKey GetTimeZoneRegistryDynamicSubKey(TimeZoneInfo tzi)
        {
            RegistryKey subKey = GetTimeZoneRegistrySubKey(tzi._id);
            return subKey.OpenSubKey("Dynamic DST", false);
        }

        public static void Sort(List<TimeZoneInfo> tzInfos)
        {
            tzInfos.Sort(new TimeZoneInfo());
        }

        public override string ToString()
        {
            return this.DisplayName;
        }

        public override bool Equals(object obj)
        {
            if (obj.GetType() == this.GetType())
            {
                TimeZoneInfo objTzi = (TimeZoneInfo)obj;
                if (this._displayName != objTzi._displayName
                    || this._id != objTzi._id
                    || !this._tzi.Equals(objTzi._tzi))
                    return false;
                else
                    return true;
            }
            return false;
        }


        public int Compare(TimeZoneInfo x, TimeZoneInfo y)
        {
            if (x._tzi.bias > y._tzi.bias)
                return -1;
            else if (x._tzi.bias < y._tzi.bias)
                return 1;
            else
                return x._displayName.CompareTo(y._displayName);
        }


        public override int GetHashCode()
        {
            return base.GetHashCode();
        }
    }
}